@@ -2,9 +2,6 @@ use crate::compiler::tokens::{Span, Token};
 use crate::error::{Error, ErrorKind};
 use crate::utils::{memchr, memstr, unescape};
 
-#[cfg(test)]
-use similar_asserts::assert_eq;
-
 #[cfg(feature = "custom_syntax")]
 pub use crate::custom_syntax::SyntaxConfig;
 
@@ -33,6 +30,7 @@ struct TokenizerState<'s> {
     failed: bool,
     current_line: u32,
     current_col: u32,
+    current_offset: u32,
 }
 
 fn find_start_marker_memchr(a: &str) -> Option<(usize, bool)> {
@@ -190,22 +188,25 @@ impl<'s> TokenizerState<'s> {
                 _ => self.current_col += 1,
             }
         }
+        self.current_offset += bytes as u32;
         self.rest = new_rest;
         skipped
     }
 
     #[inline(always)]
-    fn loc(&self) -> (u32, u32) {
-        (self.current_line, self.current_col)
+    fn loc(&self) -> (u32, u32, u32) {
+        (self.current_line, self.current_col, self.current_offset)
     }
 
-    fn span(&self, start: (u32, u32)) -> Span {
-        let (start_line, start_col) = start;
+    fn span(&self, start: (u32, u32, u32)) -> Span {
+        let (start_line, start_col, start_offset) = start;
         Span {
             start_line,
             start_col,
+            start_offset,
             end_line: self.current_line,
             end_col: self.current_col,
+            end_offset: self.current_offset,
         }
     }
 
@@ -337,6 +338,7 @@ pub fn tokenize(
         failed: false,
         current_line: 1,
         current_col: 0,
+        current_offset: 0,
     };
     let mut trim_leading_whitespace = false;
 
@@ -399,20 +401,28 @@ pub fn tokenize(
                             // raw blocks require some special handling.  If we are at the beginning of a raw
                             // block we want to skip everything until {% endraw %} completely ignoring iterior
                             // syntax and emit the entire raw block as TemplateData.
-                            if let Some((mut ptr, _)) =
+                            if let Some((raw, trim_start)) =
                                 skip_basic_tag(&state.rest[skip..], "raw", block_end)
                             {
-                                ptr += skip;
+                                state.advance(raw + skip);
+                                let mut ptr = 0;
                                 while let Some(block) =
                                     memstr(&state.rest.as_bytes()[ptr..], block_start.as_bytes())
                                 {
                                     ptr += block + block_start.len();
-                                    if let Some((endraw, trim)) =
+                                    let trim_end = state.rest.as_bytes().get(ptr) == Some(&b'-');
+                                    if let Some((endraw, trim_next)) =
                                         skip_basic_tag(&state.rest[ptr..], "endraw", block_end)
                                     {
-                                        let result = &state.rest[..ptr + endraw];
+                                        let mut result = &state.rest[..ptr - block_start.len()];
+                                        if trim_start {
+                                            result = result.trim_start();
+                                        }
+                                        if trim_end {
+                                            result = result.trim_end();
+                                        }
                                         state.advance(ptr + endraw);
-                                        trim_leading_whitespace = trim;
+                                        trim_leading_whitespace = trim_next;
                                         return Some(Ok((
                                             Token::TemplateData(result),
                                             state.span(old_loc),
@@ -568,93 +578,100 @@ pub fn tokenize(
     })
 }
 
-#[test]
-fn test_find_marker() {
-    let syntax = SyntaxConfig::default();
-    assert!(find_start_marker("{", &syntax).is_none());
-    assert!(find_start_marker("foo", &syntax).is_none());
-    assert!(find_start_marker("foo {", &syntax).is_none());
-    assert_eq!(find_start_marker("foo {{", &syntax), Some((4, false)));
-    assert_eq!(find_start_marker("foo {{-", &syntax), Some((4, true)));
-}
+#[cfg(test)]
+mod tests {
+    use super::*;
 
-#[test]
-#[cfg(feature = "custom_syntax")]
-fn test_find_marker_custom_syntax() {
-    use crate::Syntax;
+    use similar_asserts::assert_eq;
 
-    let syntax = Syntax {
-        block_start: "%{".into(),
-        block_end: "}%".into(),
-        variable_start: "[[".into(),
-        variable_end: "]]".into(),
-        comment_start: "/*".into(),
-        comment_end: "*/".into(),
-    };
-
-    let syntax_config = syntax.compile().expect("failed to create syntax config");
-
-    assert_eq!(find_start_marker("%{", &syntax_config), Some((0, false)));
-    assert!(find_start_marker("/", &syntax_config).is_none());
-    assert!(find_start_marker("foo [", &syntax_config).is_none());
-    assert_eq!(
-        find_start_marker("foo /*", &syntax_config),
-        Some((4, false))
-    );
-    assert_eq!(
-        find_start_marker("foo [[-", &syntax_config),
-        Some((4, true))
-    );
-}
-
-#[test]
-fn test_is_basic_tag() {
-    assert_eq!(skip_basic_tag(" raw %}", "raw", "%}"), Some((7, false)));
-    assert_eq!(skip_basic_tag(" raw %}", "endraw", "%}"), None);
-    assert_eq!(skip_basic_tag("  raw  %}", "raw", "%}"), Some((9, false)));
-    assert_eq!(skip_basic_tag("-  raw  -%}", "raw", "%}"), Some((11, true)));
-}
-
-#[test]
-fn test_basic_identifiers() {
-    fn assert_ident(s: &str) {
-        match tokenize(s, true, Default::default()).next() {
-            Some(Ok((Token::Ident(ident), _))) if ident == s => {}
-            _ => panic!("did not get a matching token result: {s:?}"),
-        }
+    #[test]
+    fn test_find_marker() {
+        let syntax = SyntaxConfig::default();
+        assert!(find_start_marker("{", &syntax).is_none());
+        assert!(find_start_marker("foo", &syntax).is_none());
+        assert!(find_start_marker("foo {", &syntax).is_none());
+        assert_eq!(find_start_marker("foo {{", &syntax), Some((4, false)));
+        assert_eq!(find_start_marker("foo {{-", &syntax), Some((4, true)));
     }
 
-    fn assert_not_ident(s: &str) {
-        let res = tokenize(s, true, Default::default()).collect::<Result<Vec<_>, _>>();
-        if let Ok(tokens) = res {
-            if let &[(Token::Ident(_), _)] = &tokens[..] {
-                panic!("got a single ident for {s:?}")
+    #[test]
+    #[cfg(feature = "custom_syntax")]
+    fn test_find_marker_custom_syntax() {
+        use crate::Syntax;
+
+        let syntax = Syntax {
+            block_start: "%{".into(),
+            block_end: "}%".into(),
+            variable_start: "[[".into(),
+            variable_end: "]]".into(),
+            comment_start: "/*".into(),
+            comment_end: "*/".into(),
+        };
+
+        let syntax_config = syntax.compile().expect("failed to create syntax config");
+
+        assert_eq!(find_start_marker("%{", &syntax_config), Some((0, false)));
+        assert!(find_start_marker("/", &syntax_config).is_none());
+        assert!(find_start_marker("foo [", &syntax_config).is_none());
+        assert_eq!(
+            find_start_marker("foo /*", &syntax_config),
+            Some((4, false))
+        );
+        assert_eq!(
+            find_start_marker("foo [[-", &syntax_config),
+            Some((4, true))
+        );
+    }
+
+    #[test]
+    fn test_is_basic_tag() {
+        assert_eq!(skip_basic_tag(" raw %}", "raw", "%}"), Some((7, false)));
+        assert_eq!(skip_basic_tag(" raw %}", "endraw", "%}"), None);
+        assert_eq!(skip_basic_tag("  raw  %}", "raw", "%}"), Some((9, false)));
+        assert_eq!(skip_basic_tag("-  raw  -%}", "raw", "%}"), Some((11, true)));
+    }
+
+    #[test]
+    fn test_basic_identifiers() {
+        fn assert_ident(s: &str) {
+            match tokenize(s, true, Default::default()).next() {
+                Some(Ok((Token::Ident(ident), _))) if ident == s => {}
+                _ => panic!("did not get a matching token result: {s:?}"),
             }
         }
-    }
 
-    assert_ident("foo_bar_baz");
-    assert_ident("_foo_bar_baz");
-    assert_ident("_42world");
-    assert_ident("_world42");
-    assert_ident("world42");
-    assert_not_ident("42world");
+        fn assert_not_ident(s: &str) {
+            let res = tokenize(s, true, Default::default()).collect::<Result<Vec<_>, _>>();
+            if let Ok(tokens) = res {
+                if let &[(Token::Ident(_), _)] = &tokens[..] {
+                    panic!("got a single ident for {s:?}")
+                }
+            }
+        }
 
-    #[cfg(feature = "unicode")]
-    {
-        assert_ident("foo");
-        assert_ident("föö");
-        assert_ident("き");
-        assert_ident("_");
-        assert_not_ident("1a");
-        assert_not_ident("a-");
-        assert_not_ident("🐍a");
-        assert_not_ident("a🐍🐍");
-        assert_ident("ᢅ");
-        assert_ident("ᢆ");
-        assert_ident("℘");
-        assert_ident("℮");
-        assert_not_ident("·");
-        assert_ident("a·");
+        assert_ident("foo_bar_baz");
+        assert_ident("_foo_bar_baz");
+        assert_ident("_42world");
+        assert_ident("_world42");
+        assert_ident("world42");
+        assert_not_ident("42world");
+
+        #[cfg(feature = "unicode")]
+        {
+            assert_ident("foo");
+            assert_ident("föö");
+            assert_ident("き");
+            assert_ident("_");
+            assert_not_ident("1a");
+            assert_not_ident("a-");
+            assert_not_ident("🐍a");
+            assert_not_ident("a🐍🐍");
+            assert_ident("ᢅ");
+            assert_ident("ᢆ");
+            assert_ident("℘");
+            assert_ident("℮");
+            assert_not_ident("·");
+            assert_ident("a·");
+        }
     }
 }
