@@ -11,31 +11,12 @@ use crate::compiler::parser::parse_expr;
 use crate::error::{attach_basic_debug_info, Error, ErrorKind};
 use crate::expression::Expression;
 use crate::output::Output;
-use crate::template::{CompiledTemplate, Template};
+use crate::template::{CompiledTemplate, CompiledTemplateRef, Template};
 use crate::utils::{AutoEscape, BTreeMapKeysDebug, UndefinedBehavior};
 use crate::value::{FunctionArgs, FunctionResult, Value};
-use crate::vm::{State, Vm};
+use crate::vm::State;
 use crate::{defaults, filters, functions, tests};
 
-type TemplateMap<'source> = BTreeMap<&'source str, Arc<CompiledTemplate<'source>>>;
-
-#[derive(Clone)]
-enum Source<'source> {
-    Borrowed(TemplateMap<'source>, SyntaxConfig),
-    #[cfg(feature = "source")]
-    Owned(crate::source::Source),
-}
-
-impl<'source> fmt::Debug for Source<'source> {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        match self {
-            Self::Borrowed(tmpls, _) => fmt::Debug::fmt(&BTreeMapKeysDebug(tmpls), f),
-            #[cfg(feature = "source")]
-            Self::Owned(arg0) => fmt::Debug::fmt(arg0, f),
-        }
-    }
-}
-
 type AutoEscapeFunc = dyn Fn(&str) -> AutoEscape + Sync + Send;
 type FormatterFunc = dyn Fn(&mut Output, &State, &Value) -> Result<(), Error> + Sync + Send;
 
@@ -47,13 +28,6 @@ type FormatterFunc = dyn Fn(&mut Output, &State, &Value) -> Result<(), Error> +
 /// The environment holds references to the source the templates were created from.
 /// This makes it very inconvenient to pass around unless the templates are static
 /// strings.
-#[cfg_attr(
-    feature = "source",
-    doc = "
-For situations where you want to load dynamic templates and share the
-environment it's recommended to turn on the `source` feature and to use the
-[`Source`](crate::source::Source) type with the environment."
-)]
 ///
 /// There are generally two ways to construct an environment:
 ///
@@ -63,10 +37,10 @@ environment it's recommended to turn on the `source` feature and to use the
 /// * [`Environment::empty`] creates a completely blank environment.
 #[derive(Clone)]
 pub struct Environment<'source> {
-    templates: Source<'source>,
+    templates: TemplateStore<'source>,
     filters: BTreeMap<Cow<'source, str>, filters::BoxedFilter>,
     tests: BTreeMap<Cow<'source, str>, tests::BoxedTest>,
-    pub(crate) globals: BTreeMap<Cow<'source, str>, Value>,
+    globals: BTreeMap<Cow<'source, str>, Value>,
     default_auto_escape: Arc<AutoEscapeFunc>,
     undefined_behavior: UndefinedBehavior,
     formatter: Arc<FormatterFunc>,
@@ -102,7 +76,7 @@ impl<'source> Environment<'source> {
     /// [`empty`](Environment::empty) method.
     pub fn new() -> Environment<'source> {
         Environment {
-            templates: Source::Borrowed(Default::default(), Default::default()),
+            templates: TemplateStore::default(),
             filters: defaults::get_builtin_filters(),
             tests: defaults::get_builtin_tests(),
             globals: defaults::get_globals(),
@@ -122,7 +96,7 @@ impl<'source> Environment<'source> {
     /// logic for auto escaping configured.
     pub fn empty() -> Environment<'source> {
         Environment {
-            templates: Source::Borrowed(Default::default(), Default::default()),
+            templates: TemplateStore::default(),
             filters: Default::default(),
             tests: Default::default(),
             globals: Default::default(),
@@ -136,53 +110,109 @@ impl<'source> Environment<'source> {
         }
     }
 
-    /// Loads a template from a string.
+    /// Loads a template from a string into the environment.
     ///
     /// The `name` parameter defines the name of the template which identifies
     /// it.  To look up a loaded template use the [`get_template`](Self::get_template)
     /// method.
     ///
+    /// ```
+    /// # use minijinja::Environment;
+    /// let mut env = Environment::new();
+    /// env.add_template("index.html", "Hello {{ name }}!").unwrap();
+    /// ```
+    ///
     /// Note that there are situations where the interface of this method is
-    /// too restrictive.  For instance the environment itself does not permit
-    /// any form of sensible dynamic template loading.
+    /// too restrictive as you need to hold on to the strings for the lifetime
+    /// of the environment.
     #[cfg_attr(
-        feature = "source",
-        doc = "To address this restriction use [`set_source`](Self::set_source)."
+        feature = "loader",
+        doc = "To address this restriction use [`add_template_owned`](Self::add_template_owned)."
     )]
     pub fn add_template(&mut self, name: &'source str, source: &'source str) -> Result<(), Error> {
-        match self.templates {
-            Source::Borrowed(ref mut map, ref syntax) => {
-                let compiled_template = ok!(CompiledTemplate::from_name_and_source_with_syntax(
-                    name,
-                    source,
-                    syntax.clone()
-                ));
-                map.insert(name, Arc::new(compiled_template));
-                Ok(())
-            }
-            #[cfg(feature = "source")]
-            Source::Owned(ref mut src) => src.add_template(name, source),
-        }
+        self.templates.insert(name, source)
+    }
+
+    /// Adds a template without without borrowing.
+    ///
+    /// This lets you place an owned [`String`] in the environment rather than the
+    /// borrowed `&str` without having to worry about lifetimes.
+    ///
+    /// ```
+    /// # use minijinja::Environment;
+    /// let mut env = Environment::new();
+    /// env.add_template_owned("index.html".to_string(), "Hello {{ name }}!".to_string()).unwrap();
+    /// ```
+    ///
+    /// **Note**: the name is a bit of a misnomer as this API also allows to borrow too as
+    /// the parameters are actually [`Cow`].
+    #[cfg(feature = "loader")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "loader")))]
+    pub fn add_template_owned<N, S>(&mut self, name: N, source: S) -> Result<(), Error>
+    where
+        N: Into<Cow<'source, str>>,
+        S: Into<Cow<'source, str>>,
+    {
+        self.templates.insert_cow(name.into(), source.into())
+    }
+
+    /// Register a template loader as source of templates.
+    ///
+    /// When a template loader is registered, the environment gains the ability
+    /// to dynamically load templates.  The loader is invoked with the name of
+    /// the template.  If this template exists `Ok(Some(template_source))` has
+    /// to be returned, otherwise `Ok(None)`.  Once a template has been loaded
+    /// it's stored on the environment.  This means the loader is only invoked
+    /// once per template name.
+    ///
+    /// For loading templates from the file system, you can use the
+    /// [`path_loader`](crate::path_loader) function.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// # use minijinja::Environment;
+    /// fn create_env() -> Environment<'static> {
+    ///     let mut env = Environment::new();
+    ///     env.set_loader(|name| {
+    ///         if name == "layout.html" {
+    ///             Ok(Some("...".into()))
+    ///         } else {
+    ///             Ok(None)
+    ///         }
+    ///     });
+    ///     env
+    /// }
+    /// ```
+    #[cfg(feature = "loader")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "loader")))]
+    pub fn set_loader<F>(&mut self, f: F)
+    where
+        F: Fn(&str) -> Result<Option<String>, Error> + Send + Sync + 'static,
+    {
+        self.templates.set_loader(f);
     }
 
     /// Removes a template by name.
     pub fn remove_template(&mut self, name: &str) {
-        match self.templates {
-            Source::Borrowed(ref mut map, _) => {
-                map.remove(name);
-            }
-            #[cfg(feature = "source")]
-            Source::Owned(ref mut source) => {
-                source.remove_template(name);
-            }
-        }
+        self.templates.remove(name);
+    }
+
+    /// Removes all stored templates.
+    ///
+    /// This method is mainly useful when combined with a loader as it causes
+    /// the loader to "reload" templates.  By calling this method one can trigger
+    /// a reload.
+    pub fn clear_templates(&mut self) {
+        self.templates.clear();
     }
 
     /// Fetches a template by name.
     ///
     /// This requires that the template has been loaded with
     /// [`add_template`](Environment::add_template) beforehand.  If the template was
-    /// not loaded an error of kind `TemplateNotFound` is returned.
+    /// not loaded an error of kind `TemplateNotFound` is returned.  If a loaded was
+    /// added to the engine this can also dynamically load templates.
     ///
     /// ```
     /// # use minijinja::{Environment, context};
@@ -191,45 +221,57 @@ impl<'source> Environment<'source> {
     /// let tmpl = env.get_template("hello.txt").unwrap();
     /// println!("{}", tmpl.render(context!{ name => "World" }).unwrap());
     /// ```
-    pub fn get_template(&self, name: &str) -> Result<Template<'_>, Error> {
-        let compiled = match &self.templates {
-            Source::Borrowed(ref map, _) => {
-                ok!(map.get(name).ok_or_else(|| Error::new_not_found(name)))
-            }
-            #[cfg(feature = "source")]
-            Source::Owned(source) => ok!(source.get_compiled_template(name)),
-        };
+    pub fn get_template(&self, name: &str) -> Result<Template<'_, '_>, Error> {
+        let compiled = ok!(self.templates.get(name));
         Ok(Template::new(
             self,
-            compiled,
+            CompiledTemplateRef::Borrowed(compiled),
             self.initial_auto_escape(name),
         ))
     }
 
-    /// Parses and renders a template from a string in one go.
+    /// Loads a template from a string.
     ///
-    /// In some cases you really only need a template to be rendered once from
-    /// a string and returned.  The internal name of the template is `<string>`.
+    /// In some cases you really only need to work with (eg: render) a template to be
+    /// rendered once only.
     ///
     /// ```
     /// # use minijinja::{Environment, context};
     /// let env = Environment::new();
-    /// let rv = env.render_str("Hello {{ name }}", context! { name => "World" });
+    /// let tmpl = env.template_from_named_str("template_name", "Hello {{ name }}").unwrap();
+    /// let rv = tmpl.render(context! { name => "World" });
     /// println!("{}", rv.unwrap());
     /// ```
+    pub fn template_from_named_str(
+        &self,
+        name: &'source str,
+        source: &'source str,
+    ) -> Result<Template<'_, 'source>, Error> {
+        Ok(Template::new(
+            self,
+            CompiledTemplateRef::Owned(Arc::new(ok!(CompiledTemplate::new(
+                name,
+                source,
+                self._syntax_config().clone(),
+            )))),
+            self.initial_auto_escape(name),
+        ))
+    }
+
+    /// Loads a template from a string, with name `<string>`.
     ///
-    /// **Note on values:** The [`Value`] type implements `Serialize` and can be
-    /// efficiently passed to render.  It does not undergo actual serialization.
-    pub fn render_str<S: Serialize>(&self, source: &str, ctx: S) -> Result<String, Error> {
-        // reduce total amount of code faling under mono morphization into
-        // this function, and share the rest in _eval.
-        self._render_str("<string>", source, Value::from_serializable(&ctx))
+    /// This is a shortcut to [`template_from_named_str`](Self::template_from_named_str)
+    /// with name set to `<string>`.
+    pub fn template_from_str(&self, source: &'source str) -> Result<Template<'_, 'source>, Error> {
+        self.template_from_named_str("<string>", source)
     }
 
     /// Parses and renders a template from a string in one go with name.
     ///
     /// Like [`render_str`](Self::render_str), but provide a name for the
-    /// template to be used instead of the default `<string>`.
+    /// template to be used instead of the default `<string>`.  This is an
+    /// alias for [`template_from_named_str`](Self::template_from_named_str) paired with
+    /// [`render`](Template::render).
     ///
     /// ```
     /// # use minijinja::{Environment, context};
@@ -237,7 +279,7 @@ impl<'source> Environment<'source> {
     /// let rv = env.render_named_str(
     ///     "template_name",
     ///     "Hello {{ name }}",
-    ///     context! { name => "World" }
+    ///     context!{ name => "World" }
     /// );
     /// println!("{}", rv.unwrap());
     /// ```
@@ -250,27 +292,23 @@ impl<'source> Environment<'source> {
         source: &str,
         ctx: S,
     ) -> Result<String, Error> {
-        // reduce total amount of code faling under mono morphization into
-        // this function, and share the rest in _eval.
-        self._render_str(name, source, Value::from_serializable(&ctx))
+        ok!(self.template_from_named_str(name, source)).render(ctx)
     }
 
-    fn _render_str(&self, name: &str, source: &str, root: Value) -> Result<String, Error> {
-        let compiled = ok!(CompiledTemplate::from_name_and_source_with_syntax(
-            name,
-            source,
-            self._syntax_config().clone()
-        ));
-        let mut rv = String::with_capacity(compiled.buffer_size_hint);
-        Vm::new(self)
-            .eval(
-                &compiled.instructions,
-                root,
-                &compiled.blocks,
-                &mut Output::with_string(&mut rv),
-                self.initial_auto_escape(name),
-            )
-            .map(|_| rv)
+    /// Parses and renders a template from a string in one go.
+    ///
+    /// In some cases you really only need a template to be rendered once from
+    /// a string and returned.  The internal name of the template is `<string>`.
+    ///
+    /// This is an alias for [`template_from_str`](Self::template_from_str) paired with
+    /// [`render`](Template::render).
+    ///
+    /// **Note on values:** The [`Value`] type implements `Serialize` and can be
+    /// efficiently passed to render.  It does not undergo actual serialization.
+    pub fn render_str<S: Serialize>(&self, source: &str, ctx: S) -> Result<String, Error> {
+        // reduce total amount of code faling under mono morphization into
+        // this function, and share the rest in _eval.
+        ok!(self.template_from_str(source)).render(ctx)
     }
 
     /// Sets a new function to select the default auto escaping.
@@ -391,6 +429,9 @@ impl<'source> Environment<'source> {
     /// (`None`).  To turn on fuel set something like `Some(50000)` which will
     /// allow 50.000 instructions to execute before running out of fuel.
     ///
+    /// To find out how much fuel is consumed, you can access the fuel levels
+    /// from the [`State`](crate::State).
+    ///
     /// Fuel consumed per-render.
     #[cfg(feature = "fuel")]
     #[cfg_attr(docsrs, doc(cfg(feature = "fuel")))]
@@ -414,11 +455,7 @@ impl<'source> Environment<'source> {
     #[cfg(feature = "custom_syntax")]
     #[cfg_attr(docsrs, doc(cfg(feature = "custom_syntax")))]
     pub fn set_syntax(&mut self, syntax: crate::custom_syntax::Syntax) -> Result<(), Error> {
-        match self.templates {
-            Source::Borrowed(_, ref mut syn) => *syn = ok!(syntax.compile()),
-            #[cfg(feature = "source")]
-            Source::Owned(ref mut source) => ok!(source.set_syntax(syntax)),
-        };
+        self.templates.syntax_config = ok!(syntax.compile());
         Ok(())
     }
 
@@ -430,49 +467,7 @@ impl<'source> Environment<'source> {
     }
 
     fn _syntax_config(&self) -> &SyntaxConfig {
-        match self.templates {
-            Source::Borrowed(_, ref syn) => syn,
-            #[cfg(feature = "source")]
-            Source::Owned(ref source) => source._syntax_config(),
-        }
-    }
-
-    /// Sets the template source for the environment.
-    ///
-    /// This helps when working with dynamically loaded templates.  The
-    /// [`Source`](crate::source::Source) is consulted by the environment to
-    /// look up templates that are requested.  The source has the capabilities
-    /// to load templates with fewer lifetime restrictions and can also
-    /// load templates dynamically at runtime as requested.
-    ///
-    /// When a source is set already loaded templates in the environment are
-    /// discarded and replaced with the templates from the source.
-    ///
-    /// For more information see [`Source`](crate::source::Source).
-    #[cfg(feature = "source")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "source")))]
-    pub fn set_source(&mut self, source: crate::source::Source) {
-        self.templates = Source::Owned(source);
-    }
-
-    /// Returns the currently set source.
-    #[cfg(feature = "source")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "source")))]
-    pub fn source(&self) -> Option<&crate::source::Source> {
-        match self.templates {
-            Source::Borrowed(..) => None,
-            Source::Owned(ref source) => Some(source),
-        }
-    }
-
-    /// Returns the currently set source as mutable reference.
-    #[cfg(feature = "source")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "source")))]
-    pub fn source_mut(&mut self) -> Option<&mut crate::source::Source> {
-        match self.templates {
-            Source::Borrowed(..) => None,
-            Source::Owned(ref mut source) => Some(source),
-        }
+        &self.templates.syntax_config
     }
 
     /// Compiles an expression.
@@ -569,6 +564,11 @@ impl<'source> Environment<'source> {
         self.globals.remove(name);
     }
 
+    /// Returns an empty [`State`] for testing purposes and similar.
+    pub fn empty_state(&self) -> State<'_, '_> {
+        State::new_for_env(self)
+    }
+
     /// Looks up a function.
     pub(crate) fn get_global(&self, name: &str) -> Option<Value> {
         self.globals.get(name).cloned()
@@ -606,3 +606,55 @@ impl<'source> Environment<'source> {
         }
     }
 }
+
+#[cfg(not(feature = "loader"))]
+mod basic_store {
+    use super::*;
+
+    #[derive(Clone, Default)]
+    pub struct BasicStore<'source> {
+        pub syntax_config: SyntaxConfig,
+        map: BTreeMap<&'source str, Arc<CompiledTemplate<'source>>>,
+    }
+
+    impl<'source> fmt::Debug for BasicStore<'source> {
+        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+            BTreeMapKeysDebug(&self.map).fmt(f)
+        }
+    }
+
+    impl<'source> BasicStore<'source> {
+        pub fn insert(&mut self, name: &'source str, source: &'source str) -> Result<(), Error> {
+            self.map.insert(
+                name,
+                Arc::new(ok!(CompiledTemplate::new(
+                    name,
+                    source,
+                    self.syntax_config.clone()
+                ))),
+            );
+            Ok(())
+        }
+
+        pub fn remove(&mut self, name: &str) {
+            self.map.remove(name);
+        }
+
+        pub fn clear(&mut self) {
+            self.map.clear();
+        }
+
+        pub fn get(&self, name: &str) -> Result<&CompiledTemplate<'source>, Error> {
+            self.map
+                .get(name)
+                .map(|x| &**x)
+                .ok_or_else(|| Error::new_not_found(name))
+        }
+    }
+}
+
+#[cfg(not(feature = "loader"))]
+use self::basic_store::BasicStore as TemplateStore;
+
+#[cfg(feature = "loader")]
+use crate::loader::LoaderStore as TemplateStore;
