@@ -72,6 +72,9 @@
 //! this module.  Note though that these functions are not to be
 //! called from Rust code as their exact interface (arguments and return types)
 //! might change from one MiniJinja version to another.
+//!
+//! Some additional filters are available in the
+//! [`minijinja-contrib`](https://crates.io/crates/minijinja-contrib) crate.
 use std::sync::Arc;
 
 use crate::error::Error;
@@ -253,16 +256,12 @@ mod builtins {
     use super::*;
 
     use crate::error::ErrorKind;
-    use crate::key::Key;
     use crate::value::{Kwargs, ValueKind, ValueRepr};
     use std::borrow::Cow;
     use std::cmp::Ordering;
     use std::fmt::Write;
     use std::mem;
 
-    #[cfg(test)]
-    use {crate::testutils::apply_filter, similar_asserts::assert_eq};
-
     /// Converts a value to uppercase.
     ///
     /// ```jinja
@@ -357,18 +356,19 @@ mod builtins {
     }
 
     fn sort_helper(a: &Value, b: &Value, case_sensitive: bool) -> Ordering {
-        if !case_sensitive && (a.kind() == ValueKind::String || b.kind() == ValueKind::String) {
-            // TODO: optional unicode support
-            return a
-                .to_string()
-                .to_ascii_lowercase()
-                .cmp(&b.to_string().to_ascii_lowercase());
+        if !case_sensitive {
+            if let (Some(a), Some(b)) = (a.as_str(), b.as_str()) {
+                #[cfg(feature = "unicode")]
+                {
+                    return unicase::UniCase::new(a).cmp(&unicase::UniCase::new(b));
+                }
+                #[cfg(not(feature = "unicode"))]
+                {
+                    return a.to_ascii_lowercase().cmp(&b.to_ascii_lowercase());
+                }
+            }
         }
-        match (Key::from_borrowed_value(a), Key::from_borrowed_value(b)) {
-            (Ok(a), Ok(b)) => a.partial_cmp(&b),
-            _ => a.partial_cmp(b),
-        }
-        .unwrap_or(Ordering::Less)
+        a.cmp(b)
     }
 
     /// Dict sorting functionality.
@@ -659,9 +659,7 @@ mod builtins {
         let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
             Error::new(ErrorKind::InvalidOperation, "cannot convert value to list").with_source(err)
         }));
-        Ok(iter
-            .min_by(|a, b| a.partial_cmp(b).unwrap_or(Ordering::Less))
-            .unwrap_or(Value::UNDEFINED))
+        Ok(iter.min().unwrap_or(Value::UNDEFINED))
     }
 
     /// Returns the largest item from the list.
@@ -670,9 +668,7 @@ mod builtins {
         let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
             Error::new(ErrorKind::InvalidOperation, "cannot convert value to list").with_source(err)
         }));
-        Ok(iter
-            .max_by(|a, b| a.partial_cmp(b).unwrap_or(Ordering::Less))
-            .unwrap_or(Value::UNDEFINED))
+        Ok(iter.max().unwrap_or(Value::UNDEFINED))
     }
 
     /// Returns the sorted version of the given list.
@@ -1142,7 +1138,8 @@ mod builtins {
         let mut rv = Vec::with_capacity(value.len().unwrap_or(0));
 
         // attribute mapping
-        let (args, kwargs) = Kwargs::from_args(&args);
+        let (args, kwargs): (&[Value], Kwargs) = crate::value::from_args(&args)?;
+
         if let Some(attr) = ok!(kwargs.get::<Option<Value>>("attribute")) {
             if !args.is_empty() {
                 return Err(Error::from(ErrorKind::TooManyArguments));
@@ -1195,133 +1192,43 @@ mod builtins {
         Ok(rv)
     }
 
-    #[test]
-    fn test_basics() {
-        fn test(a: u32, b: u32) -> Result<u32, Error> {
-            Ok(a + b)
-        }
+    /// Returns a list of unique items from the given iterable.
+    ///
+    /// Returns a list of unique items from the given iterable.
+    ///
+    /// ```jinja
+    /// {{ ['foo', 'bar', 'foobar', 'FooBar']|unique|list }}
+    ///   -> ['foo', 'bar', 'foobar']
+    /// ```
+    ///
+    /// The unique items are yielded in the same order as their first occurrence
+    /// in the iterable passed to the filter.  The filter will not detect
+    /// duplicate objects or arrays, only primitives such as strings or numbers.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    #[cfg(feature = "builtins")]
+    pub fn unique(values: Vec<Value>) -> Value {
+        use std::collections::BTreeSet;
 
-        let mut env = crate::Environment::new();
-        env.add_filter("test", test);
-        assert_eq!(
-            apply_filter(&env, "test", &[Value::from(23), Value::from(42)]).unwrap(),
-            Value::from(65)
-        );
-    }
+        let mut rv = Vec::new();
+        let mut seen = BTreeSet::new();
 
-    #[test]
-    fn test_rest_args() {
-        fn sum(val: u32, rest: crate::value::Rest<u32>) -> u32 {
-            rest.iter().fold(val, |a, b| a + b)
-        }
-
-        let mut env = crate::Environment::new();
-        env.add_filter("sum", sum);
-        assert_eq!(
-            apply_filter(
-                &env,
-                "sum",
-                &[
-                    Value::from(1),
-                    Value::from(2),
-                    Value::from(3),
-                    Value::from(4)
-                ][..]
-            )
-            .unwrap(),
-            Value::from(1 + 2 + 3 + 4)
-        );
-    }
-
-    #[test]
-    fn test_optional_args() {
-        fn add(val: u32, a: u32, b: Option<u32>) -> Result<u32, Error> {
-            // ensure we really get our value as first argument
-            assert_eq!(val, 23);
-            let mut sum = val + a;
-            if let Some(b) = b {
-                sum += b;
+        for item in values {
+            if !seen.contains(&item) {
+                rv.push(item.clone());
+                seen.insert(item);
             }
-            Ok(sum)
         }
 
-        let mut env = crate::Environment::new();
-        env.add_filter("add", add);
-        assert_eq!(
-            apply_filter(&env, "add", &[Value::from(23), Value::from(42)][..]).unwrap(),
-            Value::from(65)
-        );
-        assert_eq!(
-            apply_filter(
-                &env,
-                "add",
-                &[Value::from(23), Value::from(42), Value::UNDEFINED][..]
-            )
-            .unwrap(),
-            Value::from(65)
-        );
-        assert_eq!(
-            apply_filter(
-                &env,
-                "add",
-                &[Value::from(23), Value::from(42), Value::from(1)][..]
-            )
-            .unwrap(),
-            Value::from(66)
-        );
+        Value::from(rv)
     }
 
-    #[test]
-    fn test_values_in_vec() {
-        fn upper(value: &str) -> String {
-            value.to_uppercase()
-        }
-
-        fn sum(value: Vec<i64>) -> i64 {
-            value.into_iter().sum::<i64>()
-        }
-
-        let mut env = crate::Environment::new();
-        env.add_filter("upper", upper);
-        env.add_filter("sum", sum);
-
-        assert_eq!(
-            apply_filter(&env, "upper", &[Value::from("Hello World!")]).unwrap(),
-            Value::from("HELLO WORLD!")
-        );
-
-        assert_eq!(
-            apply_filter(
-                &env,
-                "sum",
-                &[Value::from(vec![Value::from(1), Value::from(2)])]
-            )
-            .unwrap(),
-            Value::from(3)
-        );
-    }
-
-    #[test]
-    fn test_seq_object_borrow() {
-        fn connect(values: &dyn crate::value::SeqObject) -> String {
-            let mut rv = String::new();
-            for item in values.iter() {
-                rv.push_str(&item.to_string())
-            }
-            rv
-        }
-
-        let mut env = crate::Environment::new();
-        env.add_filter("connect", connect);
-        assert_eq!(
-            apply_filter(
-                &env,
-                "connect",
-                &[Value::from(vec![Value::from("HELLO"), Value::from(42)])]
-            )
-            .unwrap(),
-            Value::from("HELLO42")
-        );
+    /// Pretty print a variable.
+    ///
+    /// This is useful for debugging as it better shows what's inside an object.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
+    #[cfg(feature = "builtins")]
+    pub fn pprint(value: &Value) -> String {
+        format!("{:#?}", value)
     }
 }
 
