@@ -5,10 +5,10 @@ use std::convert::TryFrom;
 use std::ops::{Deref, DerefMut};
 
 use crate::error::{Error, ErrorKind};
-use crate::key::{Key, StaticKey};
 use crate::utils::UndefinedBehavior;
 use crate::value::{
-    Arc, MapType, Object, Packed, SeqObject, StringType, Value, ValueKind, ValueMap, ValueRepr,
+    Arc, KeyRef, MapType, Object, Packed, SeqObject, StringType, Value, ValueKind, ValueMap,
+    ValueRepr,
 };
 use crate::vm::State;
 
@@ -63,11 +63,10 @@ pub trait FunctionArgs<'a> {
 /// implementing [`Object::call_method`](crate::value::Object::call_method).
 ///
 /// ```
-/// use minijinja::value::from_args;
+/// # use minijinja::value::from_args;
 /// # use minijinja::value::Value;
 /// # fn foo() -> Result<(), minijinja::Error> {
 /// # let args = vec![Value::from("foo"), Value::from(42i64)]; let args = &args[..];
-///
 /// // args is &[Value]
 /// let (string, num): (&str, i64) = from_args(args)?;
 /// # Ok(()) } fn main() { foo().unwrap(); }
@@ -75,6 +74,19 @@ pub trait FunctionArgs<'a> {
 ///
 /// Note that only value conversions are supported which means that `&State` is not
 /// a valid conversion type.
+///
+/// You can also use this function to split positional and keyword arguments ([`Kwargs`]):
+///
+/// ```
+/// # use minijinja::value::{Value, Rest, Kwargs, from_args};
+/// # use minijinja::Error;
+/// # fn foo() -> Result<(), minijinja::Error> {
+/// # let args = vec![Value::from("foo"), Value::from(42i64)]; let args = &args[..];
+/// // args is &[Value], kwargs is Kwargs
+/// let (args, kwargs): (&[Value], Kwargs) = from_args(args)?;
+/// # Ok(())
+/// # } fn main() { foo().unwrap(); }
+/// ```
 #[inline(always)]
 pub fn from_args<'a, Args>(values: &'a [Value]) -> Result<Args, Error>
 where
@@ -165,6 +177,12 @@ pub trait ArgType<'a> {
     ) -> Result<(Self::Output, usize), Error> {
         Self::from_state_and_value(state, values.get(offset))
     }
+
+    #[doc(hidden)]
+    #[inline(always)]
+    fn is_trailing() -> bool {
+        false
+    }
 }
 
 macro_rules! tuple_impls {
@@ -174,19 +192,40 @@ macro_rules! tuple_impls {
         {
             type Output = ($($name::Output,)* $rest_name::Output ,);
 
-            fn from_values(state: Option<&'a State>, values: &'a [Value]) -> Result<Self::Output, Error> {
+            fn from_values(state: Option<&'a State>, mut values: &'a [Value]) -> Result<Self::Output, Error> {
                 #![allow(non_snake_case, unused)]
+                $( let $name; )*
+                let mut $rest_name = None;
                 let mut idx = 0;
+
+                // special case: the last type is marked trailing (eg: for Kwargs) and we have at
+                // least one value.  In that case we need to read it first before going to the rest
+                // of the arguments.  This is needed to support from_args::<(&[Value], Kwargs)>
+                // or similar.
+                let rest_first = $rest_name::is_trailing() && !values.is_empty();
+                if rest_first {
+                    let (val, offset) = ok!($rest_name::from_state_and_values(state, values, values.len() - 1));
+                    $rest_name = Some(val);
+                    values = &values[..values.len() - offset];
+                }
                 $(
-                    let ($name, offset) = ok!($name::from_state_and_value(state, values.get(idx)));
+                    let (val, offset) = ok!($name::from_state_and_values(state, values, idx));
+                    $name = val;
                     idx += offset;
                 )*
-                let ($rest_name, offset) = ok!($rest_name::from_state_and_values(state, values, idx));
-                idx += offset;
+
+                if !rest_first {
+                    let (val, offset) = ok!($rest_name::from_state_and_values(state, values, idx));
+                    $rest_name = Some(val);
+                    idx += offset;
+                }
+
                 if values.get(idx).is_some() {
                     Err(Error::from(ErrorKind::TooManyArguments))
                 } else {
-                    Ok(( $($name,)* $rest_name,))
+                    // SAFETY: this is safe because both no batter what `rest_first` is set to
+                    // either way the variable is set.
+                    Ok(($($name,)* unsafe { $rest_name.unwrap_unchecked() },))
                 }
             }
         }
@@ -228,14 +267,14 @@ impl<'a> From<&'a [u8]> for Value {
 impl<'a> From<&'a str> for Value {
     #[inline(always)]
     fn from(val: &'a str) -> Self {
-        ValueRepr::String(Arc::new(val.into()), StringType::Normal).into()
+        ValueRepr::String(Arc::from(val.to_string()), StringType::Normal).into()
     }
 }
 
 impl From<String> for Value {
     #[inline(always)]
     fn from(val: String) -> Self {
-        ValueRepr::String(Arc::new(val), StringType::Normal).into()
+        ValueRepr::String(Arc::from(val), StringType::Normal).into()
     }
 }
 
@@ -256,44 +295,29 @@ impl From<()> for Value {
     }
 }
 
-impl<'a> From<Key<'a>> for Value {
-    fn from(val: Key) -> Self {
-        match val {
-            Key::Bool(val) => val.into(),
-            Key::I64(val) => val.into(),
-            Key::Char(val) => val.into(),
-            Key::String(val) => ValueRepr::String(val, StringType::Normal).into(),
-            Key::Str(val) => val.into(),
-        }
-    }
-}
-
 impl<V: Into<Value>> FromIterator<V> for Value {
     fn from_iter<T: IntoIterator<Item = V>>(iter: T) -> Self {
-        let vec = iter.into_iter().map(|v| v.into()).collect();
-
-        ValueRepr::Seq(Arc::new(vec)).into()
+        ValueRepr::Seq(Arc::new(iter.into_iter().map(Into::into).collect())).into()
     }
 }
 
-impl<K: Into<StaticKey>, V: Into<Value>> FromIterator<(K, V)> for Value {
+impl<K: Into<Value>, V: Into<Value>> FromIterator<(K, V)> for Value {
     fn from_iter<T: IntoIterator<Item = (K, V)>>(iter: T) -> Self {
         let map = iter
             .into_iter()
-            .map(|(k, v)| (k.into(), v.into()))
+            .map(|(k, v)| (KeyRef::Value(k.into()), v.into()))
             .collect();
-
         ValueRepr::Map(Arc::new(map), MapType::Normal).into()
     }
 }
 
-impl<K: Into<StaticKey>, V: Into<Value>> From<BTreeMap<K, V>> for Value {
+impl<K: Into<Value>, V: Into<Value>> From<BTreeMap<K, V>> for Value {
     fn from(val: BTreeMap<K, V>) -> Self {
         val.into_iter().map(|(k, v)| (k.into(), v.into())).collect()
     }
 }
 
-impl<K: Into<StaticKey>, V: Into<Value>> From<HashMap<K, V>> for Value {
+impl<K: Into<Value>, V: Into<Value>> From<HashMap<K, V>> for Value {
     fn from(val: HashMap<K, V>) -> Self {
         val.into_iter().map(|(k, v)| (k.into(), v.into())).collect()
     }
@@ -311,8 +335,8 @@ impl<T: Object> From<Arc<T>> for Value {
     }
 }
 
-impl From<Arc<String>> for Value {
-    fn from(value: Arc<String>) -> Self {
+impl From<Arc<str>> for Value {
+    fn from(value: Arc<str>) -> Self {
         Value(ValueRepr::String(value, StringType::Normal))
     }
 }
@@ -342,6 +366,13 @@ impl From<u128> for Value {
     }
 }
 
+impl From<char> for Value {
+    #[inline(always)]
+    fn from(val: char) -> Self {
+        ValueRepr::String(Arc::from(val.to_string()), StringType::Normal).into()
+    }
+}
+
 value_from!(bool, Bool);
 value_from!(u8, U64);
 value_from!(u16, U64);
@@ -353,7 +384,6 @@ value_from!(i32, I64);
 value_from!(i64, I64);
 value_from!(f32, F64);
 value_from!(f64, F64);
-value_from!(char, Char);
 value_from!(Arc<Vec<u8>>, Bytes);
 value_from!(Arc<Vec<Value>>, Seq);
 value_from!(Arc<dyn Object>, Dynamic);
@@ -426,7 +456,12 @@ primitive_try_from!(bool, {
     ValueRepr::Bool(val) => val,
 });
 primitive_try_from!(char, {
-    ValueRepr::Char(val) => val,
+    ValueRepr::String(ref val, _) => {
+        let mut char_iter = val.chars();
+        ok!(char_iter.next().filter(|_| char_iter.next().is_none()).ok_or_else(|| {
+            unsupported_conversion(ValueKind::String, "non single character string")
+        }))
+    },
 });
 primitive_try_from!(f32, {
     ValueRepr::U64(val) => val as f32,
@@ -456,6 +491,31 @@ impl<'a> ArgType<'a> for &str {
     }
 }
 
+impl TryFrom<Value> for Arc<str> {
+    type Error = Error;
+
+    fn try_from(value: Value) -> Result<Self, Self::Error> {
+        match value.0 {
+            ValueRepr::String(x, _) => Ok(x),
+            _ => Err(Error::new(
+                ErrorKind::InvalidOperation,
+                "value is not a string",
+            )),
+        }
+    }
+}
+
+impl<'a> ArgType<'a> for Arc<str> {
+    type Output = Arc<str>;
+
+    fn from_value(value: Option<&'a Value>) -> Result<Self::Output, Error> {
+        match value {
+            Some(value) => TryFrom::try_from(value.clone()),
+            None => Err(Error::from(ErrorKind::MissingArgument)),
+        }
+    }
+}
+
 impl<'a> ArgType<'a> for &[u8] {
     type Output = &'a [u8];
 
@@ -514,7 +574,7 @@ impl<'a> ArgType<'a> for Cow<'_, str> {
     fn from_value(value: Option<&'a Value>) -> Result<Cow<'a, str>, Error> {
         match value {
             Some(value) => Ok(match value.0 {
-                ValueRepr::String(ref s, _) => Cow::Borrowed(s.as_str()),
+                ValueRepr::String(ref s, _) => Cow::Borrowed(s as &str),
                 _ => Cow::Owned(value.to_string()),
             }),
             None => Err(Error::from(ErrorKind::MissingArgument)),
@@ -534,6 +594,27 @@ impl<'a> ArgType<'a> for &Value {
     }
 }
 
+impl<'a> ArgType<'a> for &[Value] {
+    type Output = &'a [Value];
+
+    #[inline(always)]
+    fn from_value(value: Option<&'a Value>) -> Result<&'a [Value], Error> {
+        match value {
+            Some(value) => Ok(std::slice::from_ref(value)),
+            None => Err(Error::from(ErrorKind::MissingArgument)),
+        }
+    }
+
+    fn from_state_and_values(
+        _state: Option<&'a State>,
+        values: &'a [Value],
+        offset: usize,
+    ) -> Result<(&'a [Value], usize), Error> {
+        let args = values.get(offset..).unwrap_or_default();
+        Ok((args, args.len()))
+    }
+}
+
 /// Utility type to capture remaining arguments.
 ///
 /// In some cases you might want to have a variadic function.  In that case
@@ -625,7 +706,32 @@ impl<'a, T: ArgType<'a, Output = T>> ArgType<'a> for Rest<T> {
 /// ```
 ///
 /// If for whatever reason you need a value again you can use [`Into`] to
-/// convert it back into a [`Value`].
+/// convert it back into a [`Value`].  This is particularly useful when performing
+/// calls into values.  To create a [`Kwargs`] object from scratch you can use
+/// [`FromIterator`]:
+///
+/// ```
+/// use minijinja::value::{Value, Kwargs};
+/// let kwargs = Kwargs::from_iter([
+///     ("foo", Value::from(true)),
+///     ("bar", Value::from(42)),
+/// ]);
+/// let value = Value::from(kwargs);
+/// assert!(value.is_kwargs());
+/// ```
+///
+/// When working with [`Rest`] you can use [`from_args`] to split all arguments into
+/// positional arguments and keyword arguments:
+///
+/// ```
+/// # use minijinja::value::{Value, Rest, Kwargs, from_args};
+/// # use minijinja::Error;
+/// fn my_func(args: Rest<Value>) -> Result<Value, Error> {
+///     let (args, kwargs) = from_args::<(&[Value], Kwargs)>(&args)?;
+///     // do something with args and kwargs
+/// # todo!()
+/// }
+/// ```
 #[derive(Debug, Clone)]
 pub struct Kwargs {
     values: Arc<ValueMap>,
@@ -660,6 +766,10 @@ impl<'a> ArgType<'a> for Kwargs {
         }
         Ok((Kwargs::new(Default::default()), 0))
     }
+
+    fn is_trailing() -> bool {
+        true
+    }
 }
 
 impl Kwargs {
@@ -670,33 +780,12 @@ impl Kwargs {
         }
     }
 
-    /// Split off kwargs from args.
-    ///
-    /// This is useful when [`Rest`] is used to consume all arguments:
-    ///
-    /// ```rust
-    /// # use minijinja::value::{Value, Rest, Kwargs};
-    /// fn my_func(args: Rest<Value>) -> Value {
-    ///     let (args, kwargs) = Kwargs::from_args(&args);
-    ///     // do something with args and kwargs
-    /// # todo!()
-    /// }
-    /// ```
-    pub fn from_args(args: &[Value]) -> (&[Value], Kwargs) {
-        if let Some(value) = args.last() {
-            if let ValueRepr::Map(ref map, MapType::Kwargs) = value.0 {
-                return (&args[..args.len() - 1], Kwargs::new(map.clone()));
-            }
-        }
-        (args, Kwargs::new(Default::default()))
-    }
-
     /// Get a single argument from the kwargs but don't mark it as used.
     pub fn peek<'a, T>(&'a self, key: &'a str) -> Result<T, Error>
     where
         T: ArgType<'a, Output = T>,
     {
-        T::from_value(self.values.get(&Key::Str(key)))
+        T::from_value(self.values.get(&KeyRef::Str(key)))
     }
 
     /// Gets a single argument from the kwargs and marks it as used.
@@ -731,7 +820,7 @@ impl Kwargs {
 
     /// Checks if a keyword argument exists.
     pub fn has(&self, key: &str) -> bool {
-        self.values.contains_key(&Key::Str(key))
+        self.values.contains_key(&KeyRef::Str(key))
     }
 
     /// Iterates over all passed keyword arguments.
@@ -761,6 +850,32 @@ impl Kwargs {
     }
 }
 
+impl FromIterator<(String, Value)> for Kwargs {
+    fn from_iter<T>(iter: T) -> Self
+    where
+        T: IntoIterator<Item = (String, Value)>,
+    {
+        Kwargs::new(Arc::new(
+            iter.into_iter()
+                .map(|(k, v)| (KeyRef::Value(Value::from(k)), v))
+                .collect(),
+        ))
+    }
+}
+
+impl<'a> FromIterator<(&'a str, Value)> for Kwargs {
+    fn from_iter<T>(iter: T) -> Self
+    where
+        T: IntoIterator<Item = (&'a str, Value)>,
+    {
+        Kwargs::new(Arc::new(
+            iter.into_iter()
+                .map(|(k, v)| (KeyRef::Value(Value::from(k)), v))
+                .collect(),
+        ))
+    }
+}
+
 impl From<Kwargs> for Value {
     fn from(value: Kwargs) -> Self {
         Value(ValueRepr::Map(value.values, MapType::Kwargs))
@@ -859,12 +974,33 @@ impl From<usize> for Value {
     }
 }
 
-#[test]
-fn test_as_f64() {
-    let v = Value::from(42u32);
-    let f: f64 = v.try_into().unwrap();
-    assert_eq!(f, 42.0);
-    let v = Value::from(42.5);
-    let f: f64 = v.try_into().unwrap();
-    assert_eq!(f, 42.5);
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn test_as_f64() {
+        let v = Value::from(42u32);
+        let f: f64 = v.try_into().unwrap();
+        assert_eq!(f, 42.0);
+        let v = Value::from(42.5);
+        let f: f64 = v.try_into().unwrap();
+        assert_eq!(f, 42.5);
+    }
+
+    #[test]
+    fn test_split_kwargs() {
+        let args = [
+            Value::from(42),
+            Value::from(true),
+            Value::from(Kwargs::from_iter([
+                ("foo", Value::from(1)),
+                ("bar", Value::from(2)),
+            ])),
+        ];
+        let (args, kwargs) = from_args::<(&[Value], Kwargs)>(&args).unwrap();
+        assert_eq!(args, &[Value::from(42), Value::from(true)]);
+        assert_eq!(kwargs.get::<Value>("foo").unwrap(), Value::from(1));
+        assert_eq!(kwargs.get::<Value>("bar").unwrap(), Value::from(2));
+    }
 }
