@@ -15,7 +15,9 @@
 //!
 //! ```
 //! # use minijinja::value::Value;
-//! let value = Value::from(42);
+//! let int_value = Value::from(42);
+//! let none_value = Value::from(());
+//! let true_value = Value::from(true);
 //! ```
 //!
 //! Or via the [`FromIterator`] trait:
@@ -29,6 +31,10 @@
 //! let value: Value = [("key", "value")].into_iter().collect();
 //! ```
 //!
+//! The special [`Undefined`](Value::UNDEFINED) value also exists but does not
+//! have a rust equivalent.  It can be created via the [`UNDEFINED`](Value::UNDEFINED)
+//! constant.
+//!
 //! MiniJinja will however create values via an indirection via [`serde`] when
 //! a template is rendered or an expression is evaluated.  This can also be
 //! triggered manually by using the [`Value::from_serializable`] method:
@@ -106,6 +112,7 @@ use std::cmp::Ordering;
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::fmt;
+use std::hash::{Hash, Hasher};
 use std::marker::PhantomData;
 use std::sync::Arc;
 
@@ -113,9 +120,9 @@ use serde::ser::{Serialize, Serializer};
 
 use crate::error::{Error, ErrorKind};
 use crate::functions;
-use crate::key::{Key, StaticKey};
 use crate::utils::OnDrop;
 use crate::value::object::{SimpleSeqObject, SimpleStructObject};
+use crate::value::ops::as_f64;
 use crate::value::serialize::transform;
 use crate::vm::State;
 
@@ -125,22 +132,22 @@ pub use crate::value::object::{Object, ObjectKind, SeqObject, SeqObjectIter, Str
 mod argtypes;
 #[cfg(feature = "deserialization")]
 mod deserialize;
+mod keyref;
 mod object;
 pub(crate) mod ops;
 mod serialize;
 
-#[cfg(test)]
-use similar_asserts::assert_eq;
+pub(crate) use crate::value::keyref::KeyRef;
 
 // We use in-band signalling to roundtrip some internal values.  This is
 // not ideal but unfortunately there is no better system in serde today.
 const VALUE_HANDLE_MARKER: &str = "\x01__minijinja_ValueHandle";
 
 #[cfg(feature = "preserve_order")]
-pub(crate) type ValueMap = indexmap::IndexMap<StaticKey, Value>;
+pub(crate) type ValueMap = indexmap::IndexMap<KeyRef<'static>, Value>;
 
 #[cfg(not(feature = "preserve_order"))]
-pub(crate) type ValueMap = std::collections::BTreeMap<StaticKey, Value>;
+pub(crate) type ValueMap = std::collections::BTreeMap<KeyRef<'static>, Value>;
 
 #[inline(always)]
 pub(crate) fn value_map_with_capacity(capacity: usize) -> ValueMap {
@@ -151,7 +158,7 @@ pub(crate) fn value_map_with_capacity(capacity: usize) -> ValueMap {
     }
     #[cfg(feature = "preserve_order")]
     {
-        ValueMap::with_capacity(capacity)
+        ValueMap::with_capacity(crate::utils::untrusted_size_hint(capacity))
     }
 }
 
@@ -184,16 +191,20 @@ pub fn serializing_for_value() -> bool {
     INTERNAL_SERIALIZATION.with(|flag| flag.get())
 }
 
-/// When key interning is enabled while the returned token is held keys are
-/// interned in a map, freed at the end of the section.
-#[cfg(feature = "key_interning")]
-pub(crate) use crate::key::key_interning::use_string_cache as value_optimization;
-
-/// Without key interning this is a dummy drop.
-#[cfg(not(feature = "key_interning"))]
+/// Enables value optimizations.
+///
+/// If `key_interning` is enabled, this turns on that feature, otherwise
+/// it becomes a noop.
 #[inline(always)]
 pub(crate) fn value_optimization() -> impl Drop {
-    OnDrop::new(|| {})
+    #[cfg(feature = "key_interning")]
+    {
+        crate::value::keyref::key_interning::use_string_cache()
+    }
+    #[cfg(not(feature = "key_interning"))]
+    {
+        OnDrop::new(|| {})
+    }
 }
 
 fn mark_internal_serialization() -> impl Drop {
@@ -220,8 +231,6 @@ pub enum ValueKind {
     Bool,
     /// The value is a number of a supported type.
     Number,
-    /// The value is a character.
-    Char,
     /// The value is a string.
     String,
     /// The value is a byte array.
@@ -234,18 +243,16 @@ pub enum ValueKind {
 
 impl fmt::Display for ValueKind {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        let ty = match *self {
+        f.write_str(match *self {
             ValueKind::Undefined => "undefined",
             ValueKind::None => "none",
             ValueKind::Bool => "bool",
             ValueKind::Number => "number",
-            ValueKind::Char => "char",
             ValueKind::String => "string",
             ValueKind::Bytes => "bytes",
             ValueKind::Seq => "sequence",
             ValueKind::Map => "map",
-        };
-        write!(f, "{ty}")
+        })
     }
 }
 
@@ -286,12 +293,11 @@ pub(crate) enum ValueRepr {
     U64(u64),
     I64(i64),
     F64(f64),
-    Char(char),
     None,
-    Invalid(Arc<String>),
+    Invalid(Arc<str>),
     U128(Packed<u128>),
     I128(Packed<i128>),
-    String(Arc<String>, StringType),
+    String(Arc<str>, StringType),
     Bytes(Arc<Vec<u8>>),
     Seq(Arc<Vec<Value>>),
     Map(Arc<ValueMap>, MapType),
@@ -301,13 +307,12 @@ pub(crate) enum ValueRepr {
 impl fmt::Debug for ValueRepr {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
-            ValueRepr::Undefined => write!(f, "Undefined"),
+            ValueRepr::Undefined => f.write_str("Undefined"),
             ValueRepr::Bool(val) => fmt::Debug::fmt(val, f),
             ValueRepr::U64(val) => fmt::Debug::fmt(val, f),
             ValueRepr::I64(val) => fmt::Debug::fmt(val, f),
             ValueRepr::F64(val) => fmt::Debug::fmt(val, f),
-            ValueRepr::Char(val) => fmt::Debug::fmt(val, f),
-            ValueRepr::None => write!(f, "None"),
+            ValueRepr::None => f.write_str("None"),
             ValueRepr::Invalid(ref val) => write!(f, "<invalid value: {}>", val),
             ValueRepr::U128(val) => fmt::Debug::fmt(&{ val.0 }, f),
             ValueRepr::I128(val) => fmt::Debug::fmt(&{ val.0 }, f),
@@ -320,6 +325,51 @@ impl fmt::Debug for ValueRepr {
     }
 }
 
+impl Hash for Value {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        match &self.0 {
+            ValueRepr::None | ValueRepr::Undefined => 0u8.hash(state),
+            ValueRepr::String(ref s, _) => s.hash(state),
+            ValueRepr::Bool(b) => b.hash(state),
+            ValueRepr::Invalid(s) => s.hash(state),
+            ValueRepr::Bytes(b) => b.hash(state),
+            ValueRepr::Seq(b) => b.hash(state),
+            ValueRepr::Map(m, _) => m.iter().for_each(|(k, v)| {
+                k.hash(state);
+                v.hash(state);
+            }),
+            ValueRepr::Dynamic(d) => match d.kind() {
+                ObjectKind::Plain => 0u8.hash(state),
+                ObjectKind::Seq(s) => s.iter().for_each(|x| x.hash(state)),
+                ObjectKind::Struct(s) => {
+                    if let Some(fields) = s.static_fields() {
+                        fields.iter().for_each(|k| {
+                            k.hash(state);
+                            s.get_field(k).hash(state);
+                        });
+                    } else {
+                        s.fields().iter().for_each(|k| {
+                            k.hash(state);
+                            s.get_field(k).hash(state);
+                        });
+                    }
+                }
+            },
+            ValueRepr::U64(_)
+            | ValueRepr::I64(_)
+            | ValueRepr::F64(_)
+            | ValueRepr::U128(_)
+            | ValueRepr::I128(_) => {
+                if let Ok(val) = i64::try_from(self.clone()) {
+                    val.hash(state)
+                } else {
+                    as_f64(self).map(|x| x.to_bits()).hash(state)
+                }
+            }
+        }
+    }
+}
+
 /// Represents a dynamically typed value in the template engine.
 #[derive(Clone)]
 pub struct Value(pub(crate) ValueRepr);
@@ -338,6 +388,15 @@ impl PartialEq for Value {
                 None => {
                     if let (Some(a), Some(b)) = (self.as_seq(), other.as_seq()) {
                         a.iter().eq(b.iter())
+                    } else if self.kind() == ValueKind::Map && other.kind() == ValueKind::Map {
+                        if self.len() != other.len() {
+                            return false;
+                        }
+                        if let Ok(mut iter) = self.try_iter() {
+                            iter.all(|x| self.get_item_opt(&x) == other.get_item_opt(&x))
+                        } else {
+                            false
+                        }
                     } else {
                         false
                     }
@@ -351,17 +410,47 @@ impl Eq for Value {}
 
 impl PartialOrd for Value {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
-        match (&self.0, &other.0) {
-            (ValueRepr::None, ValueRepr::None) => Some(Ordering::Equal),
-            (ValueRepr::Undefined, ValueRepr::Undefined) => Some(Ordering::Equal),
-            (ValueRepr::Bytes(a), ValueRepr::Bytes(b)) => a.partial_cmp(b),
+        Some(self.cmp(other))
+    }
+}
+
+fn f64_total_cmp(left: f64, right: f64) -> Ordering {
+    // this is taken from f64::total_cmp on newer rust versions
+    let mut left = left.to_bits() as i64;
+    let mut right = right.to_bits() as i64;
+    left ^= (((left >> 63) as u64) >> 1) as i64;
+    right ^= (((right >> 63) as u64) >> 1) as i64;
+    left.cmp(&right)
+}
+
+impl Ord for Value {
+    fn cmp(&self, other: &Self) -> Ordering {
+        let value_ordering = match (&self.0, &other.0) {
+            (ValueRepr::None, ValueRepr::None) => Ordering::Equal,
+            (ValueRepr::Undefined, ValueRepr::Undefined) => Ordering::Equal,
+            (ValueRepr::String(ref a, _), ValueRepr::String(ref b, _)) => a.cmp(b),
+            (ValueRepr::Bytes(a), ValueRepr::Bytes(b)) => a.cmp(b),
             _ => match ops::coerce(self, other) {
-                Some(ops::CoerceResult::F64(a, b)) => a.partial_cmp(&b),
-                Some(ops::CoerceResult::I128(a, b)) => a.partial_cmp(&b),
-                Some(ops::CoerceResult::Str(a, b)) => a.partial_cmp(b),
-                None => None,
+                Some(ops::CoerceResult::F64(a, b)) => f64_total_cmp(a, b),
+                Some(ops::CoerceResult::I128(a, b)) => a.cmp(&b),
+                Some(ops::CoerceResult::Str(a, b)) => a.cmp(b),
+                None => {
+                    if let (Some(a), Some(b)) = (self.as_seq(), other.as_seq()) {
+                        a.iter().cmp(b.iter())
+                    } else if self.kind() == ValueKind::Map && other.kind() == ValueKind::Map {
+                        if let (Ok(a), Ok(b)) = (self.try_iter(), other.try_iter()) {
+                            a.map(|k| (k.clone(), self.get_item_opt(&k)))
+                                .cmp(b.map(|k| (k.clone(), other.get_item_opt(&k))))
+                        } else {
+                            Ordering::Equal
+                        }
+                    } else {
+                        Ordering::Equal
+                    }
+                }
             },
-        }
+        };
+        value_ordering.then((self.kind() as usize).cmp(&(other.kind() as usize)))
     }
 }
 
@@ -375,12 +464,12 @@ impl fmt::Display for Value {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match &self.0 {
             ValueRepr::Undefined => Ok(()),
-            ValueRepr::Bool(val) => write!(f, "{val}"),
-            ValueRepr::U64(val) => write!(f, "{val}"),
-            ValueRepr::I64(val) => write!(f, "{val}"),
+            ValueRepr::Bool(val) => val.fmt(f),
+            ValueRepr::U64(val) => val.fmt(f),
+            ValueRepr::I64(val) => val.fmt(f),
             ValueRepr::F64(val) => {
                 if val.is_nan() {
-                    write!(f, "NaN")
+                    f.write_str("NaN")
                 } else if val.is_infinite() {
                     write!(f, "{}inf", if val.is_sign_negative() { "-" } else { "" })
                 } else {
@@ -391,31 +480,30 @@ impl fmt::Display for Value {
                     write!(f, "{num}")
                 }
             }
-            ValueRepr::Char(val) => write!(f, "{val}"),
-            ValueRepr::None => write!(f, "none"),
+            ValueRepr::None => f.write_str("none"),
             ValueRepr::Invalid(ref val) => write!(f, "<invalid value: {}>", val),
             ValueRepr::I128(val) => write!(f, "{}", { val.0 }),
             ValueRepr::String(val, _) => write!(f, "{val}"),
             ValueRepr::Bytes(val) => write!(f, "{}", String::from_utf8_lossy(val)),
             ValueRepr::Seq(values) => {
-                ok!(write!(f, "["));
+                ok!(f.write_str("["));
                 for (idx, val) in values.iter().enumerate() {
                     if idx > 0 {
-                        ok!(write!(f, ", "));
+                        ok!(f.write_str(", "));
                     }
                     ok!(write!(f, "{val:?}"));
                 }
-                write!(f, "]")
+                f.write_str("]")
             }
             ValueRepr::Map(m, _) => {
-                ok!(write!(f, "{{"));
+                ok!(f.write_str("{"));
                 for (idx, (key, val)) in m.iter().enumerate() {
                     if idx > 0 {
-                        ok!(write!(f, ", "));
+                        ok!(f.write_str(", "));
                     }
                     ok!(write!(f, "{key:?}: {val:?}"));
                 }
-                write!(f, "}}")
+                f.write_str("}")
             }
             ValueRepr::U128(val) => write!(f, "{}", { val.0 }),
             ValueRepr::Dynamic(x) => write!(f, "{x}"),
@@ -436,22 +524,23 @@ impl Default for Value {
 /// same functionality.  There is no guarantee that a string will be interned
 /// as there are heuristics involved for it.  Additionally the string interning
 /// will only work during the template engine execution (eg: within filters etc.).
-///
-/// ```
-/// use minijinja::value::{intern, Value};
-/// let val = Value::from(intern("my_key"));
-/// ```
-pub fn intern(s: &str) -> Arc<String> {
-    if let Key::String(ref s) = Key::make_string_key(s) {
-        s.clone()
-    } else {
-        unreachable!()
+pub fn intern(s: &str) -> Arc<str> {
+    #[cfg(feature = "key_interning")]
+    {
+        crate::value::keyref::key_interning::try_intern(s)
+    }
+    #[cfg(not(feature = "key_interning"))]
+    {
+        Arc::from(s.to_string())
     }
 }
 
 #[allow(clippy::len_without_is_empty)]
 impl Value {
-    /// The undefined value
+    /// The undefined value.
+    ///
+    /// This constant exists because the undefined type does not exist in Rust
+    /// and this is the only way to construct it.
     pub const UNDEFINED: Value = Value(ValueRepr::Undefined);
 
     /// Creates a value from something that can be serialized.
@@ -494,7 +583,7 @@ impl Value {
     /// let val = Value::from_safe_string("<em>note</em>".into());
     /// ```
     pub fn from_safe_string(value: String) -> Value {
-        ValueRepr::String(Arc::new(value), StringType::Safe).into()
+        ValueRepr::String(Arc::from(value), StringType::Safe).into()
     }
 
     /// Creates a value from a dynamic object.
@@ -589,7 +678,6 @@ impl Value {
             ValueRepr::Undefined => ValueKind::Undefined,
             ValueRepr::Bool(_) => ValueKind::Bool,
             ValueRepr::U64(_) | ValueRepr::I64(_) | ValueRepr::F64(_) => ValueKind::Number,
-            ValueRepr::Char(_) => ValueKind::Char,
             ValueRepr::None => ValueKind::None,
             ValueRepr::I128(_) => ValueKind::Number,
             ValueRepr::String(..) => ValueKind::String,
@@ -636,7 +724,6 @@ impl Value {
             ValueRepr::I64(x) => x != 0,
             ValueRepr::I128(x) => x.0 != 0,
             ValueRepr::F64(x) => x != 0.0,
-            ValueRepr::Char(x) => x != '\x00',
             ValueRepr::String(ref x, _) => !x.is_empty(),
             ValueRepr::Bytes(ref x) => !x.is_empty(),
             ValueRepr::None | ValueRepr::Undefined | ValueRepr::Invalid(_) => false,
@@ -668,7 +755,7 @@ impl Value {
     /// If the value is a string, return it.
     pub fn as_str(&self) -> Option<&str> {
         match &self.0 {
-            ValueRepr::String(ref s, _) => Some(s.as_str()),
+            ValueRepr::String(ref s, _) => Some(s as &str),
             _ => None,
         }
     }
@@ -756,7 +843,7 @@ impl Value {
     pub fn get_attr(&self, key: &str) -> Result<Value, Error> {
         Ok(match self.0 {
             ValueRepr::Undefined => return Err(Error::from(ErrorKind::UndefinedError)),
-            ValueRepr::Map(ref items, _) => items.get(&Key::Str(key)).cloned(),
+            ValueRepr::Map(ref items, _) => items.get(&KeyRef::Str(key)).cloned(),
             ValueRepr::Dynamic(ref dy) => match dy.kind() {
                 ObjectKind::Struct(s) => s.get_field(key),
                 ObjectKind::Plain | ObjectKind::Seq(_) => None,
@@ -774,7 +861,7 @@ impl Value {
     /// also not be created.
     pub(crate) fn get_attr_fast(&self, key: &str) -> Option<Value> {
         match self.0 {
-            ValueRepr::Map(ref items, _) => items.get(&Key::Str(key)).cloned(),
+            ValueRepr::Map(ref items, _) => items.get(&KeyRef::Str(key)).cloned(),
             ValueRepr::Dynamic(ref dy) => match dy.kind() {
                 ObjectKind::Struct(s) => s.get_field(key),
                 ObjectKind::Plain | ObjectKind::Seq(_) => None,
@@ -885,7 +972,7 @@ impl Value {
     }
 
     pub(crate) fn get_item_opt(&self, key: &Value) -> Option<Value> {
-        let key = some!(Key::from_borrowed_value(key).ok());
+        let key = KeyRef::Value(key.clone());
 
         let seq = match self.0 {
             ValueRepr::Map(ref items, _) => return items.get(&key).cloned(),
@@ -893,16 +980,31 @@ impl Value {
             ValueRepr::Dynamic(ref dy) => match dy.kind() {
                 ObjectKind::Plain => return None,
                 ObjectKind::Seq(s) => s,
-                ObjectKind::Struct(s) => match key {
-                    Key::String(ref key) => return s.get_field(key),
-                    Key::Str(key) => return s.get_field(key),
-                    _ => return None,
-                },
+                ObjectKind::Struct(s) => {
+                    return if let Some(key) = key.as_str() {
+                        s.get_field(key)
+                    } else {
+                        None
+                    };
+                }
             },
+            ValueRepr::String(ref s, _) => {
+                if let Some(idx) = key.as_i64() {
+                    let idx = some!(isize::try_from(idx).ok());
+                    let idx = if idx < 0 {
+                        some!(s.chars().count().checked_sub(-idx as usize))
+                    } else {
+                        idx as usize
+                    };
+                    return s.chars().nth(idx).map(Value::from);
+                } else {
+                    return None;
+                }
+            }
             _ => return None,
         };
 
-        if let Key::I64(idx) = key {
+        if let Some(idx) = key.as_i64() {
             let idx = some!(isize::try_from(idx).ok());
             let idx = if idx < 0 {
                 some!(seq.item_count().checked_sub(-idx as usize))
@@ -916,7 +1018,38 @@ impl Value {
     }
 
     /// Calls the value directly.
-    pub(crate) fn call(&self, state: &State, args: &[Value]) -> Result<Value, Error> {
+    ///
+    /// If the value holds a function or macro, this invokes it.  Note that in
+    /// MiniJinja there is a separate namespace for methods on objects and callable
+    /// items.  To call methods (which should be a rather rare occurrence) you
+    /// have to use [`call_method`](Self::call_method).
+    ///
+    /// The `args` slice is for the arguments of the function call.  To pass
+    /// keyword arguments use the [`Kwargs`](crate::value::Kwargs) type.
+    ///
+    /// Usually the state is already available when it's useful to call this method,
+    /// but when it's not available you can get a fresh template state straight
+    /// from the [`Template`](crate::Template) via [`new_state`](crate::Template::new_state).
+    ///
+    /// ```
+    /// # use minijinja::{Environment, value::{Value, Kwargs}};
+    /// # let mut env = Environment::new();
+    /// # env.add_template("foo", "").unwrap();
+    /// # let tmpl = env.get_template("foo").unwrap();
+    /// # let state = tmpl.new_state(); let state = &state;
+    /// let func = Value::from_function(|v: i64, kwargs: Kwargs| {
+    ///     v * kwargs.get::<i64>("mult").unwrap_or(1)
+    /// });
+    /// let rv = func.call(
+    ///     state,
+    ///     &[
+    ///         Value::from(42),
+    ///         Value::from(Kwargs::from_iter([("mult", Value::from(2))])),
+    ///     ],
+    /// ).unwrap();
+    /// assert_eq!(rv, Value::from(84));
+    /// ```
+    pub fn call(&self, state: &State, args: &[Value]) -> Result<Value, Error> {
         if let ValueRepr::Dynamic(ref dy) = self.0 {
             dy.call(state, args)
         } else {
@@ -928,16 +1061,14 @@ impl Value {
     }
 
     /// Calls a method on the value.
-    pub(crate) fn call_method(
-        &self,
-        state: &State,
-        name: &str,
-        args: &[Value],
-    ) -> Result<Value, Error> {
+    ///
+    /// The name of the method is `name`, the arguments passed are in the `args`
+    /// slice.
+    pub fn call_method(&self, state: &State, name: &str, args: &[Value]) -> Result<Value, Error> {
         match self.0 {
             ValueRepr::Dynamic(ref dy) => return dy.call_method(state, name, args),
             ValueRepr::Map(ref map, _) => {
-                if let Some(value) = map.get(&Key::Str(name)) {
+                if let Some(value) = map.get(&KeyRef::Str(name)) {
                     return value.call(state, args);
                 }
             }
@@ -949,25 +1080,6 @@ impl Value {
         ))
     }
 
-    pub(crate) fn try_into_key(self) -> Result<StaticKey, Error> {
-        match self.0 {
-            ValueRepr::Bool(val) => Ok(Key::Bool(val)),
-            ValueRepr::U64(v) => TryFrom::try_from(v)
-                .map(Key::I64)
-                .map_err(|_| ErrorKind::NonKey.into()),
-            ValueRepr::U128(v) => TryFrom::try_from(v.0)
-                .map(Key::I64)
-                .map_err(|_| ErrorKind::NonKey.into()),
-            ValueRepr::I64(v) => Ok(Key::I64(v)),
-            ValueRepr::I128(v) => TryFrom::try_from(v.0)
-                .map(Key::I64)
-                .map_err(|_| ErrorKind::NonKey.into()),
-            ValueRepr::Char(c) => Ok(Key::Char(c)),
-            ValueRepr::String(ref s, _) => Ok(Key::String(s.clone())),
-            _ => Err(ErrorKind::NonKey.into()),
-        }
-    }
-
     /// Iterates over the value without holding a reference.
     pub(crate) fn try_iter_owned(&self) -> Result<OwnedValueIterator, Error> {
         let (iter_state, len) = match self.0 {
@@ -1063,7 +1175,6 @@ impl Serialize for Value {
             ValueRepr::U64(u) => serializer.serialize_u64(u),
             ValueRepr::I64(i) => serializer.serialize_i64(i),
             ValueRepr::F64(f) => serializer.serialize_f64(f),
-            ValueRepr::Char(c) => serializer.serialize_char(c),
             ValueRepr::None | ValueRepr::Undefined | ValueRepr::Invalid(_) => {
                 serializer.serialize_unit()
             }
@@ -1101,7 +1212,7 @@ impl Serialize for Value {
                     } else {
                         for k in s.fields() {
                             let v = s.get_field(&k).unwrap_or(Value::UNDEFINED);
-                            ok!(map.serialize_entry(k.as_str(), &v));
+                            ok!(map.serialize_entry(&*k as &str, &v));
                         }
                     }
                     map.end()
@@ -1156,13 +1267,13 @@ impl fmt::Debug for OwnedValueIterator {
 
 enum ValueIteratorState {
     Empty,
-    Chars(usize, Arc<String>),
+    Chars(usize, Arc<str>),
     Seq(usize, Arc<Vec<Value>>),
     StaticStr(usize, &'static [&'static str]),
-    ArcStr(usize, Vec<Arc<String>>),
+    ArcStr(usize, Vec<Arc<str>>),
     DynSeq(usize, Arc<dyn Object>),
     #[cfg(not(feature = "preserve_order"))]
-    Map(Option<StaticKey>, Arc<ValueMap>),
+    Map(Option<KeyRef<'static>>, Arc<ValueMap>),
     #[cfg(feature = "preserve_order")]
     Map(usize, Arc<ValueMap>),
 }
@@ -1172,7 +1283,7 @@ impl ValueIteratorState {
         match self {
             ValueIteratorState::Empty => None,
             ValueIteratorState::Chars(offset, ref s) => {
-                s.as_str()[*offset..].chars().next().map(|c| {
+                (s as &str)[*offset..].chars().next().map(|c| {
                     *offset += c.len_utf8();
                     Value::from(c)
                 })
@@ -1205,13 +1316,13 @@ impl ValueIteratorState {
             #[cfg(feature = "preserve_order")]
             ValueIteratorState::Map(idx, map) => map.get_index(*idx).map(|x| {
                 *idx += 1;
-                Value::from(x.0.clone())
+                x.0.as_value()
             }),
             #[cfg(not(feature = "preserve_order"))]
             ValueIteratorState::Map(ptr, map) => {
                 if let Some(current) = ptr.take() {
                     let next = map.range(&current..).nth(1).map(|x| x.0.clone());
-                    let rv = Value::from(current);
+                    let rv = current.as_value();
                     *ptr = next;
                     Some(rv)
                 } else {
@@ -1222,50 +1333,67 @@ impl ValueIteratorState {
     }
 }
 
-#[test]
-fn test_dynamic_object_roundtrip() {
-    use std::sync::atomic::{self, AtomicUsize};
+#[cfg(test)]
+mod tests {
+    use super::*;
 
-    #[derive(Debug)]
-    struct X(AtomicUsize);
+    use similar_asserts::assert_eq;
 
-    impl fmt::Display for X {
-        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-            write!(f, "{}", self.0.load(atomic::Ordering::Relaxed))
-        }
-    }
+    #[test]
+    fn test_dynamic_object_roundtrip() {
+        use std::sync::atomic::{self, AtomicUsize};
 
-    impl Object for X {
-        fn kind(&self) -> ObjectKind<'_> {
-            ObjectKind::Struct(self)
-        }
-    }
+        #[derive(Debug)]
+        struct X(AtomicUsize);
 
-    impl crate::value::object::StructObject for X {
-        fn get_field(&self, name: &str) -> Option<Value> {
-            match name {
-                "value" => Some(Value::from(self.0.load(atomic::Ordering::Relaxed))),
-                _ => None,
+        impl fmt::Display for X {
+            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+                write!(f, "{}", self.0.load(atomic::Ordering::Relaxed))
             }
         }
 
-        fn static_fields(&self) -> Option<&'static [&'static str]> {
-            Some(&["value"][..])
+        impl Object for X {
+            fn kind(&self) -> ObjectKind<'_> {
+                ObjectKind::Struct(self)
+            }
         }
+
+        impl crate::value::object::StructObject for X {
+            fn get_field(&self, name: &str) -> Option<Value> {
+                match name {
+                    "value" => Some(Value::from(self.0.load(atomic::Ordering::Relaxed))),
+                    _ => None,
+                }
+            }
+
+            fn static_fields(&self) -> Option<&'static [&'static str]> {
+                Some(&["value"][..])
+            }
+        }
+
+        let x = Arc::new(X(Default::default()));
+        let x_value = Value::from(x.clone());
+        x.0.fetch_add(42, atomic::Ordering::Relaxed);
+        let x_clone = Value::from_serializable(&x_value);
+        x.0.fetch_add(23, atomic::Ordering::Relaxed);
+
+        assert_eq!(x_value.to_string(), "65");
+        assert_eq!(x_clone.to_string(), "65");
     }
 
-    let x = Arc::new(X(Default::default()));
-    let x_value = Value::from(x.clone());
-    x.0.fetch_add(42, atomic::Ordering::Relaxed);
-    let x_clone = Value::from_serializable(&x_value);
-    x.0.fetch_add(23, atomic::Ordering::Relaxed);
+    #[test]
+    fn test_string_char() {
+        let val = Value::from('a');
+        assert_eq!(char::try_from(val).unwrap(), 'a');
+        let val = Value::from("a");
+        assert_eq!(char::try_from(val).unwrap(), 'a');
+        let val = Value::from("wat");
+        assert!(char::try_from(val).is_err());
+    }
 
-    assert_eq!(x_value.to_string(), "65");
-    assert_eq!(x_clone.to_string(), "65");
-}
-
-#[test]
-#[cfg(target_pointer_width = "64")]
-fn test_sizes() {
-    assert_eq!(std::mem::size_of::<Value>(), 24);
+    #[test]
+    #[cfg(target_pointer_width = "64")]
+    fn test_sizes() {
+        assert_eq!(std::mem::size_of::<Value>(), 24);
+    }
 }
