@@ -1,4 +1,4 @@
-use std::collections::{BTreeMap, BTreeSet};
+use std::collections::BTreeMap;
 use std::mem;
 use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::{Arc, Mutex};
@@ -9,8 +9,10 @@ use crate::compiler::instructions::{
 use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::output::{CaptureMode, Output};
-use crate::utils::{AutoEscape, UndefinedBehavior};
-use crate::value::{ops, value_map_with_capacity, value_optimization, MapType, Value, ValueRepr};
+use crate::utils::{untrusted_size_hint, AutoEscape, UndefinedBehavior};
+use crate::value::{
+    ops, value_map_with_capacity, value_optimization, KeyRef, MapType, Value, ValueRepr,
+};
 use crate::vm::context::{Context, Frame, LoopState, Stack};
 use crate::vm::loop_object::Loop;
 use crate::vm::state::BlockStack;
@@ -43,9 +45,9 @@ pub struct Vm<'env> {
     env: &'env Environment<'env>,
 }
 
-fn prepare_blocks<'env, 'vm>(
-    blocks: &'vm BTreeMap<&'env str, Instructions<'env>>,
-) -> BTreeMap<&'env str, BlockStack<'vm, 'env>> {
+pub(crate) fn prepare_blocks<'env, 'template>(
+    blocks: &'template BTreeMap<&'env str, Instructions<'env>>,
+) -> BTreeMap<&'env str, BlockStack<'template, 'env>> {
     blocks
         .iter()
         .map(|(name, instr)| (*name, BlockStack::new(instr)))
@@ -74,39 +76,18 @@ impl<'env> Vm<'env> {
         Vm { env }
     }
 
-    /// Evaluates the given inputs
-    pub fn eval(
+    /// Evaluates the given inputs.
+    ///
+    /// It returns both the last value left on the stack as well as the state
+    /// at the end of the evaluation.
+    pub fn eval<'template>(
         &self,
-        instructions: &Instructions<'env>,
+        instructions: &'template Instructions<'env>,
         root: Value,
-        blocks: &BTreeMap<&'env str, Instructions<'env>>,
+        blocks: &'template BTreeMap<&'env str, Instructions<'env>>,
         out: &mut Output,
         auto_escape: AutoEscape,
-    ) -> Result<Option<Value>, Error> {
-        let _guard = value_optimization();
-        self.eval_state(
-            &mut State::new(
-                self.env,
-                Context::new(ok!(Frame::new_checked(root))),
-                auto_escape,
-                instructions,
-                prepare_blocks(blocks),
-            ),
-            out,
-        )
-    }
-
-    #[cfg(feature = "multi_template")]
-    /// Evaluates the given inputs and block
-    pub fn eval_block(
-        &self,
-        block: &str,
-        instructions: &Instructions<'env>,
-        root: Value,
-        blocks: &BTreeMap<&'env str, Instructions<'env>>,
-        out: &mut Output,
-        auto_escape: AutoEscape,
-    ) -> Result<Option<Value>, Error> {
+    ) -> Result<(Option<Value>, State<'template, 'env>), Error> {
         let _guard = value_optimization();
         let mut state = State::new(
             self.env,
@@ -115,10 +96,7 @@ impl<'env> Vm<'env> {
             instructions,
             prepare_blocks(blocks),
         );
-        out.begin_capture(CaptureMode::Discard);
-        self.eval_state(&mut state, out)?;
-        out.end_capture(state.auto_escape);
-        self.call_block(block, &mut state, out)
+        self.eval_state(&mut state, out).map(|x| (x, state))
     }
 
     /// Evaluate a macro in a state.
@@ -144,11 +122,10 @@ impl<'env> Vm<'env> {
                 env: self.env,
                 ctx,
                 current_block: None,
-                current_call: None,
                 auto_escape: state.auto_escape(),
                 instructions,
                 blocks: BTreeMap::default(),
-                loaded_templates: BTreeSet::new(),
+                loaded_templates: Default::default(),
                 #[cfg(feature = "macros")]
                 macros: state.macros.clone(),
                 #[cfg(feature = "fuel")]
@@ -340,8 +317,8 @@ impl<'env> Vm<'env> {
                     let mut map = value_map_with_capacity(*pair_count);
                     for _ in 0..*pair_count {
                         let value = stack.pop();
-                        let key = ctx_ok!(stack.pop().try_into_key());
-                        map.insert(key, value);
+                        let key = stack.pop();
+                        map.insert(KeyRef::Value(key), value);
                     }
                     stack.push(Value(ValueRepr::Map(map.into(), MapType::Normal)))
                 }
@@ -349,13 +326,13 @@ impl<'env> Vm<'env> {
                     let mut map = value_map_with_capacity(*pair_count);
                     for _ in 0..*pair_count {
                         let value = stack.pop();
-                        let key = stack.pop().try_into_key().unwrap();
-                        map.insert(key, value);
+                        let key = stack.pop();
+                        map.insert(KeyRef::Value(key), value);
                     }
                     stack.push(Value(ValueRepr::Map(map.into(), MapType::Kwargs)))
                 }
                 Instruction::BuildList(count) => {
-                    let mut v = Vec::with_capacity(*count);
+                    let mut v = Vec::with_capacity(untrusted_size_hint(*count));
                     for _ in 0..*count {
                         v.push(stack.pop());
                     }
@@ -514,7 +491,6 @@ impl<'env> Vm<'env> {
                     stack.push(out.end_capture(state.auto_escape));
                 }
                 Instruction::ApplyFilter(name, arg_count, local_id) => {
-                    state.current_call = Some(name);
                     let filter =
                         ctx_ok!(get_or_lookup_local(&mut loaded_filters, *local_id, || {
                             state.env.get_filter(name)
@@ -529,10 +505,8 @@ impl<'env> Vm<'env> {
                     a = ctx_ok!(filter.apply_to(state, args));
                     stack.drop_top(*arg_count);
                     stack.push(a);
-                    state.current_call = Some(name);
                 }
                 Instruction::PerformTest(name, arg_count, local_id) => {
-                    state.current_call = Some(name);
                     let test = ctx_ok!(get_or_lookup_local(&mut loaded_tests, *local_id, || {
                         state.env.get_test(name)
                     })
@@ -543,11 +517,8 @@ impl<'env> Vm<'env> {
                     let rv = ctx_ok!(test.perform(state, args));
                     stack.drop_top(*arg_count);
                     stack.push(Value::from(rv));
-                    state.current_call = None;
                 }
                 Instruction::CallFunction(name, arg_count) => {
-                    state.current_call = Some(name);
-
                     // super is a special function reserved for super-ing into blocks.
                     if *name == "super" {
                         if *arg_count != 0 {
@@ -578,16 +549,12 @@ impl<'env> Vm<'env> {
                             format!("{name} is unknown"),
                         ));
                     }
-
-                    state.current_call = None;
                 }
                 Instruction::CallMethod(name, arg_count) => {
-                    state.current_call = Some(name);
                     let args = stack.slice_top(*arg_count);
                     a = ctx_ok!(args[0].call_method(state, name, &args[1..]));
                     stack.drop_top(*arg_count);
                     stack.push(a);
-                    state.current_call = None;
                 }
                 Instruction::CallObject(arg_count) => {
                     let args = stack.slice_top(*arg_count);
@@ -602,13 +569,9 @@ impl<'env> Vm<'env> {
                     stack.pop();
                 }
                 Instruction::FastSuper => {
-                    // Note that we don't store 'current_call' here since it
-                    // would only be visible (and unused) internally.
                     ctx_ok!(self.perform_super(state, out, false));
                 }
                 Instruction::FastRecurse => {
-                    // Note that we don't store 'current_call' here since it
-                    // would only be visible (and unused) internally.
                     recurse_loop!(false);
                 }
                 // Explanation on the behavior of `LoadBlocks` and rendering of
@@ -653,10 +616,10 @@ impl<'env> Vm<'env> {
                 }
                 #[cfg(feature = "multi_template")]
                 Instruction::ExportLocals => {
-                    let locals = state.ctx.current_locals();
+                    let locals = state.ctx.current_locals_mut();
                     let mut module = value_map_with_capacity(locals.len());
                     for (key, value) in locals.iter() {
-                        module.insert((*key).into(), value.clone());
+                        module.insert(KeyRef::Value(Value::from(*key)), value.clone());
                     }
                     stack.push(Value(ValueRepr::Map(module.into(), MapType::Normal)));
                 }
@@ -715,9 +678,11 @@ impl<'env> Vm<'env> {
                     continue;
                 }
             };
+
+            let (new_instructions, new_blocks) = ok!(tmpl.instructions_and_blocks());
             let old_escape = mem::replace(&mut state.auto_escape, tmpl.initial_auto_escape());
-            let old_instructions = mem::replace(&mut state.instructions, tmpl.instructions());
-            let old_blocks = mem::replace(&mut state.blocks, prepare_blocks(tmpl.blocks()));
+            let old_instructions = mem::replace(&mut state.instructions, new_instructions);
+            let old_blocks = mem::replace(&mut state.blocks, prepare_blocks(new_blocks));
             let old_closure = state.ctx.take_closure();
             ok!(state.ctx.incr_depth(INCLUDE_RECURSION_COST));
             let rv = self.eval_state(state, out);
@@ -834,27 +799,27 @@ impl<'env> Vm<'env> {
             ));
         }
         let tmpl = ok!(self.env.get_template(name));
-        state.loaded_templates.insert(tmpl.instructions().name());
-        for (name, instr) in tmpl.blocks().iter() {
+        let (new_instructions, new_blocks) = ok!(tmpl.instructions_and_blocks());
+        state.loaded_templates.insert(new_instructions.name());
+        for (name, instr) in new_blocks.iter() {
             state
                 .blocks
                 .entry(name)
                 .or_insert_with(BlockStack::default)
                 .append_instructions(instr);
         }
-        Ok(tmpl.instructions())
+        Ok(new_instructions)
     }
 
     #[cfg(feature = "multi_template")]
-    fn call_block(
+    pub(crate) fn call_block(
         &self,
-        name: &'env str,
+        name: &str,
         state: &mut State<'_, 'env>,
         out: &mut Output,
     ) -> Result<Option<Value>, Error> {
-        let old_block = state.current_block;
-        state.current_block = Some(name);
-        if let Some(block_stack) = state.blocks.get(name) {
+        if let Some((name, block_stack)) = state.blocks.get_key_value(name) {
+            let old_block = mem::replace(&mut state.current_block, Some(name));
             let old_instructions =
                 mem::replace(&mut state.instructions, block_stack.instructions());
             state.ctx.push_frame(Frame::default())?;
@@ -978,7 +943,7 @@ impl<'env> Vm<'env> {
         Arc::make_mut(&mut state.macros).push((state.instructions, offset));
         stack.push(Value::from_object(Macro {
             data: Arc::new(MacroData {
-                name: Arc::new(name.to_string()),
+                name: Arc::from(name.to_string()),
                 arg_spec,
                 macro_ref_id,
                 closure,
